from pyorderbook import Book, ask, bid


def test_bid() -> None:
    book = Book()
    blotter = book.match([ask("IBM", 3.5, 70), ask("IBM", 3.6, 70), bid("IBM", 54.3, 140)])
    assert len(blotter[2].trades) == 2
    assert blotter[2].average_price == 3.55


def test_ask() -> None:
    book = Book()
    blotters = book.match([bid("GOOG", 3.5, 70), bid("GOOG", 3.6, 70), ask("GOOG", 54.3, 140)])
    assert len(blotters[2].trades) == 0
    assert blotters[2].average_price == 0
    blotter = book.match(ask("GOOG", 3.1, 140))
    assert len(blotter.trades) == 2
    assert blotter.average_price == 3.55


def test_cancel() -> None:
    book = Book()
    bid1, bid2 = bid("GOOG", 3.5, 70), bid("GOOG", 3.6, 70)
    book.match([bid1, bid2])
    book.cancel(bid1)
    blotter = book.match(ask("GOOG", 3.1, 140))
    assert len(blotter.trades) == 1
    assert blotter.average_price == 3.6



from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("pyorderbook")
except PackageNotFoundError:
    __version__ = "v0.4.6"

from pyorderbook.book import Book
from pyorderbook.level import PriceLevel
from pyorderbook.order import Order, OrderQueue, OrderStatus, Side, ask, bid
from pyorderbook.trade_blotter import Trade, TradeBlotter

easter_egg = "artificial lake"
__all__ = [
    "Book",
    "bid",
    "ask",
    "Order",
    "OrderQueue",
    "OrderStatus",
    "Side",
    "PriceLevel",
    "Trade",
    "TradeBlotter",
    "easter_egg",
]



from pyorderbook import Book, ask, bid


def simulate_order_flow() -> None:
    """Toy order flow simulation. Feel free to experiment and set
    your desired log level at the top of the file!"""
    book = Book()
    book.match(ask("GOOG", 3.6, 70))
    book.match(ask("GOOG", 3.5, 70))
    blotter = book.match(bid("GOOG", 3.7, 70))
    print(blotter)


if __name__ == "__main__":
    simulate_order_flow()



from dataclasses import dataclass, field
from decimal import Decimal
from typing import TypeAlias
from uuid import UUID

from pyorderbook.order import Order

Price: TypeAlias = Decimal


@dataclass
class Trade:
    """Stores trade data for matched orders"""

    incoming_order_id: UUID
    standing_order_id: UUID
    fill_quantity: int
    fill_price: Price


@dataclass
class TradeBlotter:
    """blotter statistics return by the Book().match function.
    Displays order status, executed trades, and order statistics
    including total cost and average price.
    """

    order: Order
    trades: list[Trade]
    num_trades: int = field(default=0, init=False)
    total_cost: float = field(default=0, init=False)
    average_price: float = field(default=0, init=False)

    def __post_init__(self) -> None:
        if self.trades:
            self.total_cost = round(
                float(sum(trade.fill_price * trade.fill_quantity for trade in self.trades)), 2
            )
            self.average_price = round(
                float(sum(trade.fill_price for trade in self.trades) / len(self.trades)), 2
            )



import heapq as pq
import logging
from collections import defaultdict
from decimal import Decimal
from typing import TypeAlias, overload
from uuid import UUID

from pyorderbook.level import PriceLevel
from pyorderbook.order import Order, Side
from pyorderbook.trade_blotter import Trade, TradeBlotter

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

Symbol: TypeAlias = str
Price: TypeAlias = Decimal
PriceLevelHeap: TypeAlias = list[PriceLevel]


class Book:
    """Main Order Book class. Orchestrates order handling and stores
    orders.
    """

    def __init__(self) -> None:
        """Creates required data structures for order matching"""
        self.levels: defaultdict[str, dict[Side, PriceLevelHeap]] = defaultdict(
            lambda: {Side.BID: [], Side.ASK: []}
        )
        self.level_map: defaultdict[str, dict[Side, dict[Price, PriceLevel]]] = defaultdict(
            lambda: {Side.BID: {}, Side.ASK: {}}
        )
        self.order_map: dict[UUID, Order] = {}

    @overload
    def match(self, orders: Order) -> TradeBlotter: ...

    @overload
    def match(self, orders: list[Order]) -> list[TradeBlotter]: ...

    def match(self, orders: Order | list[Order]) -> TradeBlotter | list[TradeBlotter]:
        """Match incoming order(s) against standing orders in the book.
        :param order: incoming order
        :returns: TradeBlotter object containing trade metadata on the trades
        which occured during the matching process
        """
        if isinstance(orders, list):
            return [self.match(order) for order in orders]
        elif isinstance(orders, Order):
            return self._match(orders)
        raise ValueError("Invalid input type", type(orders))

    def _match(self, incoming_order: Order) -> TradeBlotter:
        """Main Order procesing function which executes the price-time priority
        matching logic.
        :param incoming_order: incoming order
        :returns: TradeBlotter object containing trade metadata on the trades
        which occured during the matching process
        """
        logger.info("~~~ Processing order: %s", incoming_order)
        trades: list[Trade] = []
        price_comparator = incoming_order.side.price_comparator
        levels = self.levels[incoming_order.symbol][incoming_order.side.other]
        while incoming_order.quantity and levels:
            level = levels[0]
            if not price_comparator(incoming_order.price, level.price):
                break
            while incoming_order.quantity and level.orders:
                best_standing_order = level.orders.peek()
                trade: Trade = self.fill(incoming_order, best_standing_order)
                trades.append(trade)
                if not best_standing_order.quantity:
                    logger.debug(
                        "Filled standing Order Id: %s from book: %s",
                        best_standing_order.id,
                    )
                    level.orders.popleft()
                    self.order_map.pop(best_standing_order.id)

            if not level.orders:
                logger.info(
                    "Flushing Price Level for %s at %s price %s",
                    incoming_order.symbol,
                    str(level.side),
                    level.price,
                )
                self.level_map[incoming_order.symbol][level.side].pop(level.price)
                pq.heappop(self.levels[incoming_order.symbol][level.side])

        if incoming_order.quantity:
            self.enqueue_order(incoming_order)

        trade_blotter = TradeBlotter(incoming_order, trades)
        logger.info("%s", trade_blotter)
        return trade_blotter

    def cancel(self, order: Order) -> None:
        """Cancel Standing Order. Remove order from its price level and delete
        reference in order id map
        :param order_id: id field of Order object
        :returns: False if order doesn't exist or if it's already cancelled,
        True if cancelled successfully.
        """
        order_id = order.id
        logger.info("~~~ Processing Cancel Request for Order Id", order_id)
        try:
            self.order_map.pop(order.id)
        except KeyError:
            logger.error("Order %s doesnt exist", order_id)
            raise

        level = self.get_level(order.symbol, order.side, order.price)
        if level is None:
            raise ValueError(
                f"Price Level {order.symbol}:{order.side}:{order.price} doesn't exist!"
            )
        level.orders.pop(order_id)

    def fill(
        self,
        incoming_order: Order,
        standing_order: Order,
    ) -> Trade:
        """Execute order fill. Updates incoming and standing order objects, determines
        trade price based on side, and matched quantity, and
        returns Trade object containing trade metadata.
        :param standing_order: standing order
        :param incoming_order: incoming order
        :returns: Trade object containing metadata on the matched price/quantity and
        order id of the matched orders
        """
        matched_quantity: int = min(standing_order.quantity, incoming_order.quantity)
        standing_order.quantity -= matched_quantity
        incoming_order.quantity -= matched_quantity
        fill_price = incoming_order.side.calc_fill_price(incoming_order.price, standing_order.price)
        trade = Trade(incoming_order.id, standing_order.id, matched_quantity, fill_price)
        logger.info("Filled Order: %s", trade)
        return trade

    def get_level(self, symbol: str, side: Side, price: Decimal) -> PriceLevel | None:
        """Return price level queue for a symbol/side/price
        :param symbol: order symbol
        :param side: order side
        :param price: order price
        :returns: price level (OrderQueue)
        """
        return self.level_map[symbol][side].get(price, None)

    def enqueue_order(self, order: Order) -> None:
        """Add order to book.
        - enqueue order to price level
        - add reference to order in order_map
        :param order: order to add to book
        :returns: None
        """
        logger.info("Adding Order to book: %s", order)
        level = self.get_level(order.symbol, order.side, order.price)
        if level is None:
            # create price level
            level = PriceLevel(order.side, order.price)
            # add level object to self.levels heap
            pq.heappush(self.levels[order.symbol][order.side], level)
            # add level reference to level map
            self.level_map[order.symbol][order.side][order.price] = level
        level.orders.append_order(order)
        self.order_map[order.id] = order

    def get_order(self, order_id: UUID) -> Order | None:
        """Return order object from order id
        :param order_id: id field of Order object
        :returns: Order object
        """
        return self.order_map.get(order_id, None)



import re
from pathlib import Path

import toml


def increment_version(version_str: str) -> str:
    major, minor, patch = map(int, version_str[1:].split("."))
    patch += 1
    if patch > 9:
        patch = 0
        minor += 1
    if minor > 9:
        minor = 0
        major += 1
    return f"v{major}.{minor}.{patch}"


def update_versions() -> str:
    repo_root = Path(__file__).parent
    pyproject_path = repo_root / "pyproject.toml"
    init_path = repo_root / "pyorderbook" / "__init__.py"

    pyproject = toml.load(pyproject_path)
    current_version = pyproject["project"]["version"]
    new_version = increment_version(current_version)

    pyproject["project"]["version"] = new_version
    with open(pyproject_path, "w") as f:
        toml.dump(pyproject, f)

    init_content = init_path.read_text()
    new_init_content = re.sub(
        r'(__version__\s*=\s*["\'].+?["\'])', f'__version__ = "{new_version}"', init_content
    )
    init_path.write_text(new_init_content)

    return new_version


if __name__ == "__main__":
    new_version = update_versions()
    print(f"Updated version to {new_version}")



from dataclasses import dataclass, field
from decimal import Decimal
from typing import TypeAlias

from pyorderbook.order import OrderQueue, Side

Price: TypeAlias = Decimal


@dataclass
class PriceLevel:
    side: Side
    price: Price
    orders: OrderQueue = field(default_factory=OrderQueue)

    def __lt__(self, other: "PriceLevel") -> bool:
        """BIDS should be max heap, asks min heap"""
        return self.side.price_comparator(self.price, other.price)



import operator
from collections.abc import Callable
from decimal import Decimal
from enum import StrEnum, auto
from functools import partial
from typing import TypeAlias
from uuid import UUID, uuid4

Symbol: TypeAlias = str
Price: TypeAlias = Decimal


class Side(StrEnum):
    """Enum to represent BID or ASK Order"""

    BID = auto()
    ASK = auto()

    @property
    def other(self) -> "Side":
        """Return the opposite side of the order"""
        return Side.BID if self == Side.ASK else Side.ASK

    @property
    def price_comparator(self) -> Callable[[Price, Price], bool]:
        """Return the price comparator function for the side.
        Best price for BID is the lowest price, and vice versa for ASK.
        """
        return operator.le if self == Side.ASK else operator.ge

    @property
    def calc_fill_price(self) -> Callable[[Price, Price], Price]:
        """Return the fill price calculation function for the side.
        For BID, the fill price is the max of the two prices.
        For ASK, the fill price is the min of the two prices.
        """
        return max if self == Side.ASK else min


class OrderStatus(StrEnum):
    """Enum for Order Status after matching occurs"""

    QUEUED = auto()
    PARTIAL_FILL = auto()
    FILLED = auto()


class Order:
    def __init__(self, side: Side, symbol: Symbol, price: float, quantity: int) -> None:
        if quantity <= 0:
            raise ValueError("Order quantity must be greater than zero")
        self.id: UUID = uuid4()
        self.price: Price = Decimal(str(price))
        self.quantity: int = quantity
        self.symbol: Symbol = symbol
        self.side: Side = side
        self.original_quantity: int = quantity

    @property
    def status(self) -> OrderStatus:
        if self.quantity == 0:
            return OrderStatus.FILLED
        elif self.quantity < self.original_quantity:
            return OrderStatus.PARTIAL_FILL
        return OrderStatus.QUEUED


bid = partial(Order, Side.BID)
ask = partial(Order, Side.ASK)


class OrderQueue(dict[UUID, Order]):
    def append_order(self, order: Order) -> None:
        self[order.id] = order

    def peek(self) -> Order:
        if not self:
            raise ValueError("Order Queue is Empty!")
        first_key = next(iter(self))
        return self[first_key]

    def popleft(self) -> None:
        self.pop(self.peek().id)
